"""RFC 3986 compliant, scheme-agnostic replacement for urlparse.
The urlparse module is not compliant with RFC 3986, and is generally
unusable with custom (private) URI schemes.

"""

from __future__ import unicode_literals

import collections
import re
import urllib

URI_RE = re.compile(r"""
(?:([^:/?#]+):)?  # scheme
(?://([^/?#]*))?  # authority
([^?#]*)          # path
(?:\?([^#]*))?    # query
(?:\#(.*))?       # fragment
""", flags=(re.VERBOSE))

URI_FIELDS = ('scheme', 'authority', 'path', 'query', 'fragment')


class SplitResult(collections.namedtuple('SplitResult', URI_FIELDS)):

    def getscheme(self):
        return urllib.unquote(self.scheme)

    def getauthority(self):
        return urllib.unquote(self.authority)

    def getpath(self):
        return urllib.unquote(self.path)

    def getquery(self):
        return urllib.unquote(self.query)

    def getfragment(self):
        return urllib.unquote(self.fragment)

    def geturi(self):
        return uriunsplit(self)


def urisplit(uri):
    """Split a URI into a named tuple with five components:
    <scheme>://<authority>/<path>?<query>#<fragment>.  Note that
    %-escapes are not expaneded.

    """
    return SplitResult(*URI_RE.match(uri).groups(''))


def uriunsplit(data):
    """Combine the elements of a tuple as returned by urisplit() into a
    complete URI as a string.  The data argument can be any five-item
    iterable.  Note that this may result in a slightly different, but
    equivalent URI string.

    """
    scheme, authority, path, query, fragment = data
    if authority:
        if path.startswith('/'):
            uri = '//' + authority + path
        else:
            uri = '//' + authority + '/' + path
    elif path.startswith('//'):
        # RFC 3986 3.3: If a URI does not contain an authority
        # component, then the path cannot begin with two slash
        # characters ("//").
        raise ValueError('invalid uri path: %s' % path)
    else:
        uri = path
    if scheme:
        uri = scheme + ':' + uri
    if query:
        uri += '?' + query
    if fragment:
        uri += '#' + fragment
    return uri


def uricompose(scheme='', authority='', path='', query='', fragment=''):
    # TODO: check quoting
    scheme = urllib.quote(scheme, '')
    authority = urllib.quote(authority, ':')
    path = urllib.quote(path, ':/')
    query = urllib.quote(query, ':/?')
    fragment = urllib.quote(fragment, ':/?#')
    return uriunsplit((scheme, authority, path, query, fragment))



from __future__ import unicode_literals

import logging
import pykka

from mopidy import backend

from .library import PodcastLibraryProvider
from .playback import PodcastPlaybackProvider

logger = logging.getLogger(__name__)


class PodcastBackend(pykka.ThreadingActor, backend.Backend):

    URI_SCHEME = 'podcast'

    uri_schemes = [URI_SCHEME]

    def __init__(self, config, audio):
        super(PodcastBackend, self).__init__()
        self.config = config
        self.library = PodcastLibraryProvider(backend=self)
        self.playback = PodcastPlaybackProvider(audio=audio, backend=self)

    def getconfig(self, name):
        return self.config[self.URI_SCHEME][name]



from __future__ import unicode_literals

from mopidy import config, ext

__version__ = '0.2.0'


class Extension(ext.Extension):

    dist_name = 'Mopidy-Podcast'
    ext_name = 'podcast'
    version = __version__

    def get_default_config(self):
        import os
        conf_file = os.path.join(os.path.dirname(__file__), 'ext.conf')
        return config.read(conf_file)

    def get_config_schema(self):
        schema = super(Extension, self).get_config_schema()
        schema['feed_urls'] = config.List()
        schema['browse_label'] = config.String()
        schema['update_interval'] = config.Integer(minimum=0)
        schema['sort_order'] = config.String(choices=['asc', 'desc'])
        schema['preload'] = config.Deprecated()  # FIXME: remove
        return schema

    def setup(self, registry):
        from .backend import PodcastBackend
        registry.add('backend', PodcastBackend)



from __future__ import unicode_literals

import collections
import logging

logger = logging.getLogger(__name__)

QUERY_FIELDS = {
    'uri',
    'track_name',
    'track_no',
    'album',
    'artist',
    'composer',
    'performer',
    'albumartist',
    'genre',
    'date',
    'comment',
    'any'
}

DEFAULT_FILTERS = dict.fromkeys(QUERY_FIELDS, lambda qv, value: False)

TRACK_FILTERS = dict(
    DEFAULT_FILTERS,
    uri=lambda qv, track: qv == track.uri,
    track_name=lambda qv, track: qv == track.name,
    track_no=lambda qv, track: str(qv).isdigit() and int(qv) == track.track_no,
    album=lambda qv, track: track.album and qv == track.album.name,
    artist=lambda qv, track: any(
        qv == a.name for a in track.artists
    ),
    composer=lambda qv, track: any(
        qv == a.name for a in track.composers
    ),
    performer=lambda qv, track: any(
        qv == a.name for a in track.performers
    ),
    albumartist=lambda qv, track: track.album and any(
        qv == a.name for a in track.album.artists
    ),
    genre=lambda qv, track: qv == track.genre,
    date=lambda qv, track: qv == track.date,
    comment=lambda qv, track: qv == track.comment
)

ALBUM_FILTERS = dict(
    DEFAULT_FILTERS,
    uri=lambda qv, album: qv == album.uri,
    album=lambda qv, album: qv == album.name,
    artist=lambda qv, album: any(
        qv == a.name for a in album.artists
    ),
    albumartist=lambda qv, album: any(
        qv == a.name for a in album.artists
    ),
    date=lambda qv, album: qv == album.date
)

ARTIST_FILTERS = dict(
    DEFAULT_FILTERS,
    uri=lambda qv, artist: qv == artist.uri,
    artist=lambda qv, artist: qv == artist.name
)


# setup 'any' filters
def _any_filter(filtermap):
    filters = [filtermap[key] for key in filtermap.keys() if key != 'any']

    def any_filter(qv, value):
        return any(f(qv, value) for f in filters)
    return any_filter

TRACK_FILTERS['any'] = _any_filter(TRACK_FILTERS)
ALBUM_FILTERS['any'] = _any_filter(ALBUM_FILTERS)
ARTIST_FILTERS['any'] = _any_filter(ARTIST_FILTERS)


class Query(collections.Mapping):

    class QV(object):

        def __init__(self, value):
            self.__value = value.strip().lower()

        def __eq__(self, other):
            return other and self.__value in other.lower()

        def __ne__(self, other):
            return not other or self.__value not in other.lower()

        def __str__(self):
            return self.__value

        def __int__(self):
            return int(self.__value)

        __hash__ = None

    def __init__(self, query, exact=False):
        self.__query = {}
        for field, values in query.iteritems():
            if field not in QUERY_FIELDS:
                raise LookupError('Invalid query field "%s"' % field)
            if not values:
                raise LookupError('Missing query value for "%s"' % field)
            if not hasattr(values, '__iter__'):
                values = [values]
            if not all(values):
                raise LookupError('Missing query value for "%s"' % field)
            if exact:
                self.__query[field] = values
            else:
                self.__query[field] = [self.QV(value) for value in values]

    def __getitem__(self, key):
        return self.__query.__getitem__(key)

    def __iter__(self):
        return self.__query.__iter__()

    def __len__(self):
        return self.__query.__len__()

    def filter_tracks(self, tracks):
        return filter(self._get_filter(TRACK_FILTERS), tracks)

    def filter_albums(self, albums):
        return filter(self._get_filter(ALBUM_FILTERS), albums)

    def filter_artists(self, artists):
        return filter(self._get_filter(ARTIST_FILTERS), artists)

    def _get_filter(self, filtermap):
        from functools import partial
        filters = []
        for field, values in self.__query.iteritems():
            filters.extend(partial(filtermap[field], qv) for qv in values)

        def filterfunc(model):
            return all(f(model) for f in filters)
        return filterfunc



from __future__ import unicode_literals

import itertools
import logging
import time

from mopidy import backend
from mopidy.models import Album, Artist, Track, Ref, SearchResult

from .podcast import Podcast
from .query import Query
from .uritools import uricompose, urisplit

logger = logging.getLogger(__name__)


class DebugTimer(object):
    def __init__(self, msg):
        self.msg = msg
        self.start = None

    def __enter__(self):
        #logger.debug('%s...', self.msg)
        self.start = time.time()

    def __exit__(self, exc_type, exc_value, traceback):
        duration = (time.time() - self.start) * 1000
        logger.debug('%s took %dms', self.msg, duration)


class PodcastLibraryProvider(backend.LibraryProvider):

    PODCAST, ALBUM, TRACKS, TIMESTAMP = (0, 1, 2, 3)

    def __init__(self, backend):
        super(PodcastLibraryProvider, self).__init__(backend)
        self.uri_scheme = self.backend.URI_SCHEME
        self.root_directory = Ref.directory(
            uri=uricompose(self.uri_scheme, path='/'),
            name=self.getconfig('browse_label')
        )

        self.podcasts = {}
        for url in self.getconfig('feed_urls'):
            try:
                self.podcasts[url] = self._load_podcast(url)
            except Exception as e:
                logger.error('Error loading podcast %s: %s', url, e)
        logger.info("Loaded %d podcasts", len(self.podcasts))

    def lookup(self, uri):
        # lookup needs to be fast, so don't update
        uriparts = urisplit(uri)
        entry = self.podcasts.get(uriparts.getpath())

        if not entry:
            return []
        elif not uriparts.fragment:
            return entry[self.TRACKS].values()
        else:
            return [entry[self.TRACKS][uriparts.getfragment()]]

    def browse(self, uri):
        self._update_podcasts()

        refs = []
        if not uri:
            refs.append(self.root_directory)
        elif uri == self.root_directory.uri:
            for album in (i[self.ALBUM] for i in self.podcasts.values()):
                refs.append(Ref.directory(uri=album.uri, name=album.name))
        else:
            url = urisplit(uri).getpath()
            for track in self.podcasts[url][self.TRACKS].values():
                refs.append(Ref.track(uri=track.uri, name=track.name))
        return refs

    def find_exact(self, query=None, uris=None):
        return self._search_podcasts(Query(query, exact=True))

    def search(self, query=None, uris=None):
        return self._search_podcasts(Query(query, exact=False))

    def getstream(self, uri):
        return urisplit(uri).getfragment()

    def getconfig(self, name):
        return self.backend.getconfig(name)

    def _load_podcast(self, url):
        with DebugTimer('Loading podcast %s' % url):
            podcast = Podcast(url)
        album = self._podcast_to_album(podcast)
        tracks = {}
        for index, e in enumerate(self._sort_episodes(podcast)):
            if not e.enclosure or not 'url' in e.enclosure:
                continue
            stream = e.enclosure['url']   # TODO: store under guid?
            kwargs = {
                'uri': uricompose(self.uri_scheme, path=url, fragment=stream),
                'name': e.title,
                'album': album,
                'genre': podcast.category,
                'track_no': index + 1,
                'comment': e.subtitle
            }
            if e.author:
                kwargs['artists'] = [Artist(name=e.author)]
            else:
                kwargs['artists'] = album.artists
            if e.pubdate:
                kwargs['date'] = e.pubdate.date().isoformat()
            if e.duration:
                kwargs['length'] = int(e.duration.total_seconds() * 1000)
            tracks[stream] = Track(**kwargs)
        return (podcast, album, tracks, time.time())

    def _update_podcasts(self):
        expired = time.time() - self.getconfig('update_interval')
        updated = 0

        with DebugTimer('Updating podcasts'):
            # check for podcasts that couldn't be loaded at startup
            for url in self.getconfig('feed_urls'):
                e = self.podcasts.get(url)
                if e and e[self.TIMESTAMP] > expired:
                    continue  # podcast up-to-date
                try:
                    self.podcasts[url] = self._load_podcast(url)
                except Exception as e:
                    logger.error('Error loading podcast %s: %s', url, e)
                    updated += 1
        if updated:
            logger.info("Updated %d podcasts", updated)

    def _search_podcasts(self, query):
        self._update_podcasts()

        tracks = [v[self.TRACKS].values() for v in self.podcasts.values()]
        albums = [v[self.ALBUM] for v in self.podcasts.values()]

        return SearchResult(
            uri=uricompose(self.uri_scheme, query=repr(query)),
            tracks=query.filter_tracks(itertools.chain(*tracks)),
            albums=query.filter_albums(albums)
        )

    def _podcast_to_album(self, podcast):
        kwargs = {
            'uri': uricompose(self.uri_scheme, path=podcast.url),
            'name': podcast.title,
            'num_tracks': len(podcast)
        }
        if podcast.author:
            kwargs['artists'] = [Artist(name=podcast.author)]
        if podcast.image and 'href' in podcast.image:
            kwargs['images'] = [podcast.image['href']]
        # TBD: date
        return Album(**kwargs)

    def _sort_episodes(self, podcast):
        reverse = self.getconfig('sort_order') == 'desc'

        def keyfunc(e):
            from datetime import datetime
            return e.pubdate if e.pubdate else datetime.min

        return sorted(podcast.episodes, key=keyfunc, reverse=reverse)



from __future__ import unicode_literals

import unittest
import pykka

from mopidy_podcast.backend import PodcastBackend
from mopidy import core


class LibraryTest(unittest.TestCase):
    config = {
        'podcast': {
            'feed_urls': [],
            'browse_label': 'Podcasts',
            'update_interval': 0,
            'sort_order': 'desc'
        }
    }

    def setUp(self):
        self.backend = PodcastBackend.start(
            config=self.config, audio=None).proxy()
        self.core = core.Core(backends=[self.backend])
        self.library = self.core.library

    def tearDown(self):
        pykka.ActorRegistry.stop_all()

    def test_search_artist(self):
        self.library.search(artist=['foo'])
        # TODO: write tests

    def test_search_album(self):
        self.library.search(album=['foo'])
        # TODO: write tests

    def test_search_date(self):
        self.library.search(date=['2014-02-01'])
        # TODO: write tests



from __future__ import unicode_literals

import logging

from mopidy import backend

logger = logging.getLogger(__name__)


class PodcastPlaybackProvider(backend.PlaybackProvider):

    def change_track(self, track):
        track = track.copy(uri=self.backend.library.getstream(track.uri))
        return super(PodcastPlaybackProvider, self).change_track(track)



from __future__ import unicode_literals
import unittest

from mopidy_podcast.backend import PodcastBackend


class BackendTest(unittest.TestCase):
    config = {
        'podcast': {
            'feed_urls': ['http://www.npr.org/rss/podcast.php?id=510019'],
            'browse_label': 'Podcasts',
            'preload': False
        }
    }

    def setUp(self):
        self.backend = PodcastBackend(config=self.config, audio=None)



"""
Podcast implementation according to
https://www.apple.com/itunes/podcasts/specs.html
"""
from __future__ import unicode_literals

import collections
import datetime
import logging
import re

DURATION_RE = re.compile(r"""
(?:
    (?:(?P<hours>\d+):)?
    (?P<minutes>\d+):
)?
(?P<seconds>\d+)
""", flags=re.VERBOSE)

NAMESPACES = {
    'itunes': 'http://www.itunes.com/dtds/podcast-1.0.dtd'
}

logger = logging.getLogger(__name__)


def _settagattr(self, name, item, tag, default=None, convert=None,
                namespaces=None):
    e = item.find(tag, namespaces=namespaces)
    if e is None:
        return setattr(self, name, default)
    elif convert is not None:
        return setattr(self, name, convert(e))
    else:
        return setattr(self, name, e.text.strip())


def _to_datetime(e):
    from email.utils import mktime_tz, parsedate_tz
    try:
        timestamp = mktime_tz(parsedate_tz(e.text))
    except AttributeError:
        logger.warn('Invalid podcast date element "%s"', e.text)
        return None
    except TypeError:
        logger.warn('Invalid podcast date element "%s"', e.text)
        return None
    return datetime.datetime.fromtimestamp(timestamp)


def _to_timedelta(e):
    try:
        groups = DURATION_RE.match(e.text).groupdict('0')
    except AttributeError:
        logger.warn('Invalid podcast duration element "%s"', e.text)
        return None
    except TypeError:
        logger.warn('Invalid podcast duration element "%s"', e.text)
        return None
    return datetime.timedelta(**{k: int(v) for k, v in groups.items()})


def _to_wordlist(e):
    return [s.strip() for s in e.text.split(',')]


class Podcast(collections.Sequence):

    def __init__(self, url):
        self.url = url
        self.update()

    def __getitem__(self, index):
        return self.episodes.__getitem__(index)

    def __iter__(self):
        return self.episodes.__iter__()

    def __len__(self):
        return self.episodes.__len__()

    def update(self):
        from urllib2 import urlopen
        import xml.etree.ElementTree as ET

        channel = ET.parse(urlopen(self.url)).find('channel')
        for name in ('title', 'link', 'description', 'language', 'copyright'):
            self._settagattr(name, channel, name)
        for name in ('author', 'complete', 'explicit', 'subtitle', 'summary'):
            self._settagattr(name, channel, 'itunes:' + name)
        self._settagattr('image', channel, 'itunes:image', lambda e: e.attrib)
        self._settagattr('keywords', channel, 'itunes:keywords', _to_wordlist)

        # self.category: first browse (top-level) category
        self._settagattr('category', channel, 'itunes:category',
                         lambda e: e.get('text'))
        # self.categories: {category: [sub-category, ...], ...}
        self.categories = {}
        for e in channel.iterfind('itunes:category', namespaces=NAMESPACES):
            subcats = e.iterfind('itunes:category', namespaces=NAMESPACES)
            self.categories[e.get('text')] = [i.get('text') for i in subcats]

        self.episodes = [Episode(item) for item in channel.iter(tag='item')]

    def _settagattr(self, name, item, tag, convert=None):
        _settagattr(self, name, item, tag, None, convert, NAMESPACES)


class Episode(object):

    def __init__(self, item):
        for name in ('title', 'link', 'description', 'guid'):
            self._settagattr(name, item, name)
        for name in ('author', 'block', 'explicit', 'subtitle', 'summary'):
            self._settagattr(name, item, 'itunes:' + name)
        self._settagattr('enclosure', item, 'enclosure', lambda e: e.attrib)
        self._settagattr('pubdate', item, 'pubDate', _to_datetime)
        self._settagattr('duration', item, 'itunes:duration', _to_timedelta)
        self._settagattr('keywords', item, 'itunes:keywords', _to_wordlist)
        self._settagattr('order', item, 'itunes:order', lambda e: int(e.text))

        # normalization/fallbacks
        if not self.guid and self.enclosure:
            self.guid = self.enclosure.get('url')

    def _settagattr(self, name, item, tag, convert=None):
        _settagattr(self, name, item, tag, None, convert, NAMESPACES)


if __name__ == '__main__':
    import argparse
    import sys

    parser = argparse.ArgumentParser()
    parser.add_argument('url', metavar='URL', help='Podcast URL')
    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('-n', '--no-episodes', action='store_true')
    parser.add_argument('-v', '--verbose', action='store_true')
    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    podcast = Podcast(args.url)

    if args.verbose:
        print '\n%12s: %s' % ('URL', podcast.url)
        for name in ('title', 'link', 'description', 'language', 'copyright'):
            print '%12s: %s' % (name.title(), getattr(podcast, name))
        for name in ('author', 'complete', 'explicit', 'subtitle', 'summary'):
            print '%12s: %s' % (name.title(), getattr(podcast, name))
        for name in ('image', 'keywords', 'category', 'categories'):
            print '%12s: %s' % (name.title(), getattr(podcast, name))
    else:
        print "%s [%s]" % (podcast.title, podcast.author)

    if args.no_episodes:
        sys.exit(0)

    for index, episode in enumerate(podcast):
        if args.verbose:
            print '\n%12s: #%d [%r]' % ('Episode', index + 1, episode.order)
            for name in ('title', 'link', 'description', 'guid'):
                print '%12s: %s' % (name.title(), getattr(episode, name))
            for name in ('author', 'block', 'explicit', 'subtitle', 'summary'):
                print '%12s: %s' % (name.title(), getattr(episode, name))
            for name in ('enclosure', 'pubdate', 'duration', 'keywords'):
                print '%12s: %s' % (name.title(), getattr(episode, name))
        else:
            print '#%d: [%s] %s' % (index + 1, episode.pubdate, episode.title)
